import cupy as cp
from autogradgpu.tensor import Tensor
from autogradgpu.parameter import Parameter
from metalgpu.module import Module
from autogradgpu.dependency import Dependency
import math
import copy
from metalgpu.layers.activation_functions import Sigmoid, ReLU, TanH, Softmax
from metalgpu.utils.layer_data_manipulations import *


activation_functions = {
    'relu': ReLU,
    'sigmoid': Sigmoid,
    'tanh': TanH,
    'softmax':Softmax,
}

class Layer(Module):
    __slots__ = ('input_shape')
    """docstring for Layer."""

    def set_input_shape(self, shape):
        """ Sets the shape that the layer expects of the input in the forward
        pass method """
        self.input_shape = shape

    def layer_name(self):
        """ The name of the layer. Used in model summary. """
        return self.__class__.__name__

    def parameters_(self):
        """ The number of trainable parameters used by the layer """
        return 0

    def forward_pass(self, X, training):
        """ Propogates the signal forward in the network """
        raise NotImplementedError()

    def update_pass(self):
        """ Propogates the accumulated gradient backwards in the network.
        If the has trainable weights then these weights are also tuned in this method.
        As input (accum_grad) it receives the gradient with respect to the output of the layer and
        returns the gradient with respect to the output of the previous layer. """
        raise NotImplementedError()

    def output_shape(self):
        """ The shape of the output produced by forward_pass """
        raise NotImplementedError()



class Activation(Layer):
    __slots__=('activation_name','activation_func','trainable','layer_input')
    """A layer that applies an activation operation to the input.

    Parameters:
    -----------
    name: string
        The name of the activation function that will be used.
    """

    def __init__(self, name):
        self.activation_name = name
        self.activation_func = activation_functions[name]()
        self.trainable = True

    def layer_name(self):
        return "Activation (%s)" % (self.activation_func.__class__.__name__)

    def forward_pass(self, X, training):
        return self.activation_func(X, training)

    def update_pass(self):
        self.zero_grad()

    def output_shape(self):
        return self.input_shape

class PoolingLayer(Layer):
    """A parent class of MaxPooling2D and AveragePooling2D
    """
    def __init__(self, pool_shape=(2, 2), stride=1, padding=0):
        self.pool_shape = pool_shape
        self.stride = stride
        self.padding = padding
        self.trainable = True

    def forward_pass(self, x, training=True):
        self.layer_input = x
        X = x.data
        self.type = type(x)

        requires_grad = x.requires_grad

        batch_size, channels, height, width = X.shape

        _, out_height, out_width = self.output_shape()

        X = X.reshape(batch_size*channels, 1, height, width)
        X_col = image_to_column(X, self.pool_shape, self.stride, self.padding)

        # MaxPool or AveragePool specific method
        output = self._pool_forward(X_col)

        output = output.reshape(out_height, out_width, batch_size, channels)
        output = output.transpose(2, 3, 0, 1)

        if training:
            if requires_grad:
                depends_on = [Dependency(x, self.grad_pool)]
            else:
                depends_on = []
        else:
            depends_on = []

        return self.type(data=output, requires_grad=requires_grad, depends_on=depends_on)

    def grad_pool(self, accum_grad):
        batch_size, _, _, _ = accum_grad.shape
        channels, height, width = self.input_shape
        accum_grad = accum_grad.transpose(2, 3, 0, 1).ravel()

        # MaxPool or AveragePool specific method
        accum_grad_col = self._pool_backward(accum_grad)

        accum_grad = column_to_image(accum_grad_col, (batch_size * channels, 1, height, width), self.pool_shape, self.stride, self.padding)
        accum_grad = accum_grad.reshape((batch_size,) + self.input_shape)

        return accum_grad

    def update_pass(self):
        pass

    def output_shape(self):
        channels, height, width = self.input_shape
        out_height = (height - self.pool_shape[0]) / self.stride + 1
        out_width = (width - self.pool_shape[1]) / self.stride + 1
        assert out_height % 1 == 0
        assert out_width % 1 == 0
        return channels, int(out_height), int(out_width)



from __future__ import division
import numpy as np
from metal.utils.data_operation import accuracy_score
from metal.layers.activation_functions import Sigmoid
from autograd.dependency import Dependency

class Loss(object):
    def loss(self, y, p):
        return NotImplementedError()

    def gradient(self, y, p):
        raise NotImplementedError()

    def acc(self, y, p):
        return 0

class SquareLoss(Loss):
    __slots__ = ('y','p','type')
    def __init__(self,y=None,p=None):
        self.y = y
        self.p = p
        self.type = type(p)

    def __call__(self,y=None,p=None):
        self.y = y
        self.p = p
        self.type = type(p)


    def loss(self):
        data = np.mean(0.5 * np.power((y.data - p.data), 2))
        requires_grad = self.p.requires_grad
        if requires_grad:
            depends_on = [Dependency(self.p, self.gradient_SquareLoss)]
        else:
            depends_on = []
        return  self.type(data=data,requires_grad=requires_grad,depends_on=depends_on)


    def gradient_SquareLoss(self,grad):
        return -(self.y - self.p)

class CrossEntropy(Loss):
    __slots__ = ('y','p','type')
    def __init__(self,y=None,p=None):
        self.y = y
        self.p = p
        self.type = type(p)

    def __call__(self,y=None,p=None):
        self.y = y
        self.p = p
        self.type = type(p)

    def loss(self, training):
        # Avoid division by zero
        p = np.clip(self.p.data.astype('float64'), 1e-15, 1 - 1e-15)
        data = np.mean(- self.y.data.astype('float64') * np.log(p) - (1 - self.y.data.astype('float64')) * np.log(1 - p))
        requires_grad = self.p.requires_grad
        if training:
            if requires_grad:
                depends_on = [Dependency(self.p, self.gradient_CrossEntropy)]
            else:
                depends_on = []
        else:
            depends_on = []
        return  self.type(data=data,requires_grad=requires_grad,depends_on=depends_on)

    def gradient_CrossEntropy(self,grad):
        # Avoid division by zero
        p = np.clip(self.p.data.astype('float64'), 1e-15, 1 - 1e-15)
        d = - (self.y.data.astype('float64') / p) + (1 - self.y.data.astype('float64')) / (1 - p) * grad.astype('float64')
        return d.astype('float32')

    def acc(self):
        return accuracy_score(np.argmax(self.y.data, axis=1), np.argmax(self.p.data, axis=1))



from typing import Iterator
import inspect
from autogradgpu.tensor import Tensor
from autogradgpu.parameter import Parameter

class Module:
    def parameters(self) -> Iterator[Parameter]:
        for name, value in inspect.getmembers(self):
            if isinstance(value, Parameter):
                yield value
            if isinstance(value, Tensor):
                yield value
            elif isinstance(value, Module):
                yield from value.parameters()

    def zero_grad(self):
        for parameter in self.parameters():
            parameter.zero_grad()
            del parameter.depends_on[:]



import cupy as cp
from autogradgpu.tensor import Tensor
from autogradgpu.parameter import Parameter
from autogradgpu.dependency import Dependency
import math
import copy
from metalgpu.layers.layer import Layer
from metalgpu.utils.layer_data_manipulations import *
import numpy as np

class Conv2D(Layer):
    __slots__ =('n_filters','filter_shape','padding','stride','trainable','seed','w','b','w_opt','b_opt','type','layer_input','X_col','W_col')

    """A 2D Convolution Layer.
    Parameters:
    -----------
    n_filters: int
        The number of filters that will convolve over the input matrix. The number of channels
        of the output shape.
    filter_shape: tuple
        A tuple (filter_height, filter_width).
    input_shape: tuple
        The shape of the expected input of the layer. (batch_size, channels, height, width)
        Only needs to be specified for first layer in the network.
    padding: string
        Either 'same' or 'valid'. 'same' results in padding being added so that the output height and width
        matches the input height and width. For 'valid' no padding is added.
    stride: int
        The stride length of the filters during the convolution over the input.
    """
    def __init__(self, n_filters, filter_shape, input_shape=None, padding='same', stride=1, seed=None):
        self.n_filters = n_filters
        self.filter_shape = filter_shape
        self.padding = padding
        self.stride = stride
        self.input_shape = input_shape
        self.trainable = True
        self.seed = seed
        self.load_params_ = False

    def load_params(self, weights, bias, trainable=False):
        self.load_params_ = True
        self.trainable = trainable

        if self.trainable == False:
            self.w = Parameter(data = weights,requires_grad=False)
            self.b = Parameter(data = bias.reshape(-1,1),requires_grad=False)
        elif self.trainable == True:
            self.w = Parameter(data = weights)
            self.b = Parameter(data = bias.reshape(-1,1))


    def load_optimzer(self,optimizer=None):
        # Weight optimizers
        if optimizer is not None:
            self.w_opt  = copy.copy(optimizer)
            self.b_opt = copy.copy(optimizer)

    def initialize(self, optimizer=None):
        cp.random.seed(self.seed)
        # Initialize the weights
        filter_height, filter_width = self.filter_shape
        channels = self.input_shape[0]
        limit = 1 / math.sqrt(cp.prod(self.filter_shape))
        # create filter
        # grad would be zeros instead of None :bugfix
        if self.trainable == False:
            self.w = Parameter(data = cp.random.uniform(-limit, limit, size=(self.n_filters, channels, filter_height, filter_width)),requires_grad=False)
            self.b = Parameter(data = cp.zeros((self.n_filters, 1)),requires_grad=False)
        elif self.trainable == True:
            self.w = Parameter(data = cp.random.uniform(-limit, limit, size=(self.n_filters, channels, filter_height, filter_width)))
            self.b = Parameter(data = cp.zeros((self.n_filters, 1)))
        # Weight optimizers
        if optimizer is not None:
            self.w_opt  = copy.copy(optimizer)
            self.b_opt = copy.copy(optimizer)

    def parameters_(self):
        return cp.prod(self.w.shape) + cp.prod(self.b.shape)

    def forward_pass(self, x, training):

        X = x.data
        self.type = type(x)
        requires_grad = x.requires_grad or self.w.requires_grad or self.b.requires_grad
        depends_on: List[Dependency] = []

        batch_size, channels, height, width = X.shape
        self.layer_input = X
        # Turn image shape into column shape
        # (enables dot product between input and weights)
        self.X_col = image_to_column(X, self.filter_shape, stride=self.stride, output_shape=self.padding)
        # Turn weights into column shape
        self.W_col = self.w.data.reshape((self.n_filters, -1))
        # Calculate output
        output = self.W_col.dot(self.X_col) + self.b.data
        # Reshape into (n_filters, out_height, out_width, batch_size)
        output = output.reshape(self.output_shape() + (batch_size, ))
        # Redistribute axises so that batch size comes first
        if training:
            if requires_grad:
                if self.w.requires_grad:
                    depends_on.append(Dependency(self.w, self.grad_w_conv2D))
                if self.b.requires_grad:
                    depends_on.append(Dependency(self.b, self.grad_b_conv2D))
                if x.requires_grad:
                    depends_on.append(Dependency(x, self.grad_a_conv2D))
            else:
                depends_on = []
        return self.type(data=output.transpose(3,0,1,2),requires_grad=requires_grad,depends_on=depends_on)


    def grad_w_conv2D(self, accum_grad):
        # Reshape accumulated gradient into column shape
        accum_grad = accum_grad.transpose(1, 2, 3, 0).reshape(self.n_filters, -1)
        # Take dot product between column shaped accum. gradient and column shape
        # layer input to determine the gradient at the layer with respect to layer weights
        grad_w = accum_grad.dot(self.X_col.T).reshape(self.w.shape)

        return grad_w

    def grad_b_conv2D(self, accum_grad):
        # Reshape accumulated gradient into column shape
        accum_grad = accum_grad.transpose(1, 2, 3, 0).reshape(self.n_filters, -1)
        # The gradient with respect to bias terms is the sum similarly to in Dense layer
        grad_w0 = cp.sum(accum_grad, axis=1, keepdims=True)

        return grad_w0

    def grad_a_conv2D(self, accum_grad):
        # Reshape accumulated gradient into column shape
        accum_grad = accum_grad.transpose(1, 2, 3, 0).reshape(self.n_filters, -1)
        # Recalculate the gradient which will be propogated back to prev. layer
        accum_grad = self.W_col.T.dot(accum_grad)
        # Reshape from column shape to image shape
        accum_grad = column_to_image(accum_grad,
                                self.layer_input.shape,
                                self.filter_shape,
                                stride=self.stride,
                                output_shape=self.padding)

        return accum_grad

    def update_pass(self):
        # Update the layer weights
        if self.trainable:
            self.w = self.w_opt.update(self.w)
            self.b = self.b_opt.update(self.b)
        # clear the gradients
        self.zero_grad()




    def output_shape(self):
        channels, height, width = self.input_shape
        pad_h, pad_w = determine_padding(self.filter_shape, output_shape=self.padding)
        pad_h = np.sum(pad_h)
        pad_w = np.sum(pad_w)
        output_height = (height + pad_h - self.filter_shape[0]) / self.stride + 1
        output_width = (width + pad_w - self.filter_shape[1]) / self.stride + 1
        return self.n_filters, int(output_height), int(output_width)



import progressbar
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.pyplot as plt
import matplotlib.cm as cmx
import matplotlib.colors as colors
import cupy as cp

from metalgpu.utils.data_operation import calculate_covariance_matrix
from metalgpu.utils.data_operation import calculate_correlation_matrix
from metalgpu.utils.data_manipulation import standardize

bar_widgets = [
    'Training: ', progressbar.Percentage(), ' ', progressbar.Bar(marker="-", left="[", right="]"),
    ' ', progressbar.ETA()
]

class Plot():
    def __init__(self):
        self.cmap = plt.get_cmap('viridis')

    def _transform(self, X, dim):
        covariance = calculate_covariance_matrix(X)
        eigenvalues, eigenvectors = cp.linalg.eig(covariance)
        # Sort eigenvalues and eigenvector by largest eigenvalues
        idx = eigenvalues.argsort()[::-1]
        eigenvalues = eigenvalues[idx][:dim]
        eigenvectors = cp.atleast_1d(eigenvectors[:, idx])[:, :dim]
        # Project the data onto principal components
        X_transformed = X.dot(eigenvectors)

        return X_transformed


    def plot_regression(self, lines, title, axis_labels=None, mse=None, scatter=None, legend={"type": "lines", "loc": "lower right"}):

        if scatter:
            scatter_plots = scatter_labels = []
            for s in scatter:
                scatter_plots += [plt.scatter(s["x"], s["y"], color=s["color"], s=s["size"])]
                scatter_labels += [s["label"]]
            scatter_plots = tuple(scatter_plots)
            scatter_labels = tuple(scatter_labels)

        for l in lines:
            li = plt.plot(l["x"], l["y"], color=s["color"], linewidth=l["width"], label=l["label"])

        if mse:
            plt.suptitle(title)
            plt.title("MSE: %.2f" % mse, fontsize=10)
        else:
            plt.title(title)

        if axis_labels:
            plt.xlabel(axis_labels["x"])
            plt.ylabel(axis_labels["y"])

        if legend["type"] == "lines":
            plt.legend(loc="lower_left")
        elif legend["type"] == "scatter" and scatter:
            plt.legend(scatter_plots, scatter_labels, loc=legend["loc"])

        plt.show()



    # Plot the dataset X and the corresponding labels y in 2D using PCA.
    def plot_in_2d(self, X, y=None, title=None, accuracy=None, legend_labels=None):
        X_transformed = self._transform(X, dim=2)
        x1 = X_transformed[:, 0]
        x2 = X_transformed[:, 1]
        class_distr = []

        y = cp.array(y).astype(int)

        colors = [self.cmap(i) for i in cp.linspace(0, 1, len(cp.unique(y)))]

        # Plot the different class distributions
        for i, l in enumerate(cp.unique(y)):
            _x1 = x1[y == l]
            _x2 = x2[y == l]
            _y = y[y == l]
            class_distr.append(plt.scatter(_x1, _x2, color=colors[i]))

        # Plot legend
        if not legend_labels is None:
            plt.legend(class_distr, legend_labels, loc=1)

        # Plot title
        if title:
            if accuracy:
                perc = 100 * accuracy
                plt.suptitle(title)
                plt.title("Accuracy: %.1f%%" % perc, fontsize=10)
            else:
                plt.title(title)

        # Axis labels
        plt.xlabel('Principal Component 1')
        plt.ylabel('Principal Component 2')

        plt.show()

    # Plot the dataset X and the corresponding labels y in 3D using PCA.
    def plot_in_3d(self, X, y=None):
        X_transformed = self._transform(X, dim=3)
        x1 = X_transformed[:, 0]
        x2 = X_transformed[:, 1]
        x3 = X_transformed[:, 2]
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        ax.scatter(x1, x2, x3, c=y)
        plt.show()



from __future__ import division
import cupy as cp
import math
import sys


def calculate_entropy(y):
    """ Calculate the entropy of label array y """
    log2 = lambda x: math.log(x) / math.log(2)
    unique_labels = cp.unique(y)
    entropy = 0
    for label in unique_labels:
        count = len(y[y == label])
        p = count / len(y)
        entropy += -p * log2(p)
    return entropy


def mean_squared_error(y_true, y_pred):
    """ Returns the mean squared error between y_true and y_pred """
    mse = cp.mean(cp.power(y_true - y_pred, 2))
    return mse


def calculate_variance(X):
    """ Return the variance of the features in dataset X """
    mean = cp.ones(cp.shape(X)) * X.mean(0)
    n_samples = cp.shape(X)[0]
    variance = (1 / n_samples) * cp.diag((X - mean).T.dot(X - mean))
    
    return variance


def calculate_std_dev(X):
    """ Calculate the standard deviations of the features in dataset X """
    std_dev = cp.sqrt(calculate_variance(X))
    return std_dev


def euclidean_distance(x1, x2):
    """ Calculates the l2 distance between two vectors """
    distance = 0
    # Squared distance between each coordinate
    for i in range(len(x1)):
        distance += pow((x1[i] - x2[i]), 2)
    return math.sqrt(distance)


def accuracy_score(y_true, y_pred):
    """ Compare y_true to y_pred and return the accuracy """
    accuracy = cp.sum(y_true == y_pred, axis=0) / len(y_true)
    return accuracy


def calculate_covariance_matrix(X, Y=None):
    """ Calculate the covariance matrix for the dataset X """
    if Y is None:
        Y = X
    n_samples = cp.shape(X)[0]
    covariance_matrix = (1 / (n_samples-1)) * (X - X.mean(axis=0)).T.dot(Y - Y.mean(axis=0))

    return cp.array(covariance_matrix, dtype=float)
 

def calculate_correlation_matrix(X, Y=None):
    """ Calculate the correlation matrix for the dataset X """
    if Y is None:
        Y = X
    n_samples = cp.shape(X)[0]
    covariance = (1 / n_samples) * (X - X.mean(0)).T.dot(Y - Y.mean(0))
    std_dev_X = cp.expand_dims(calculate_std_dev(X), 1)
    std_dev_y = cp.expand_dims(calculate_std_dev(Y), 1)
    correlation_matrix = cp.divide(covariance, std_dev_X.dot(std_dev_y.T))

    return cp.array(correlation_matrix, dtype=float)



import cupy as cp
from autogradgpu.tensor import Tensor
from autogradgpu.parameter import Parameter
from metalgpu.module import Module
from autogradgpu.dependency import Dependency
import math
import copy
from metalgpu.layers.layer import Layer
import numpy as np

class Flatten(Layer):
    __slots__ =( 'prev_shape', 'trainable','type' )

    """ Turns a multidimensional matrix into two-dimensional """
    def __init__(self, input_shape=None):
        self.prev_shape = None
        self.trainable = True
        self.input_shape = input_shape

    def forward_pass(self, x_, training=True):
        self.prev_shape = x_.shape
        X = x_.data
        self.type = type(x_)
        requires_grad = x_.requires_grad
        if training:
            if requires_grad:
                depends_on = [Dependency(x_, self.gardflatten)]
            else:
                depends_on = []
        else:
            depends_on = []
        return self.type(data=X.reshape((X.shape[0], -1)),requires_grad=requires_grad,depends_on=depends_on)

    def gardflatten(self, accum_grad):
        return accum_grad.reshape(self.prev_shape)

    def update_pass(self):
        pass

    def output_shape(self):
        return (np.prod(self.input_shape),)



import numpy as np
from autograd.tensor import Tensor
from autograd.parameter import Parameter
from metal.module import Module
from autograd.dependency import Dependency
import math
import copy
from metal.layers.layer import Layer
from metal.utils.layer_data_manipulations import *

class Dense(Layer):
    __slots__ =( 'layer_input','n_units','trainable', 'w','b', 'seed','type' )
    """A fully-connected NN layer.
    Parameters:
    -----------
    n_units: int
        The number of neurons in the layer.
    input_shape: tuple
        The expected input shape of the layer. For dense layers a single digit specifying
        the number of features of the input. Must be specified if it is the first layer in
        the network.
    example shape:
        w = np.random.randn(6,5)
        i = np.random.randn(2,6)
        np.dot(i,w) + np.random.randn(1,5)
    """

    def __init__(self, n_units, input_shape=None, seed=None):
        self.layer_input = None
        self.input_shape = input_shape
        self.n_units = n_units
        self.trainable = True
        self.w = None
        self.b = None
        self.seed = seed
        self.load_params_ = False

    def load_params(self, weights, bias, trainable = False):
        self.load_params_ = True
        self.trainable = trainable
        if self.trainable == False:
            self.w = Parameter(data = weights,requires_grad=False)
            self.b = Parameter(data = bias.reshape(-1,1),requires_grad=False)
        elif self.trainable == True:
            self.w = Parameter(data = weights)
            self.b = Parameter(data = bias.reshape(-1,1))


    def load_optimzer(self,optimizer=None):
        # Weight optimizers
        if optimizer is not None:
            self.w_opt  = copy.copy(optimizer)
            self.b_opt = copy.copy(optimizer)

    def initialize(self, optimizer=None):
        np.random.seed(self.seed)
        # Initialize the weights
        limit = 1 / math.sqrt(self.input_shape[0])

        if self.trainable == False:
            self.w = Parameter(data = np.random.uniform(-limit, limit, (self.input_shape[0], self.n_units)), requires_grad=False)
            self.b = Parameter(data = np.zeros((1, self.n_units)),requires_grad=False)
        elif self.trainable == True:
            self.w = Parameter(data = np.random.uniform(-limit, limit, (self.input_shape[0], self.n_units)))
            self.b = Parameter(data = np.zeros((1, self.n_units)))
        # Weight optimizers
        if optimizer is not None:
            self.w_opt  = copy.copy(optimizer)
            self.b_opt = copy.copy(optimizer)

    def parameters_(self):
        return np.prod(self.W.shape) + np.prod(self.b.shape)

    def forward_pass(self, x, training=True):
        self.layer_input = x.data
        self.type = type(x)
        depends_on: List[Dependency] = []

        data = x.data.dot(self.w.data) + self.b.data
        requires_grad = x.requires_grad or self.w.requires_grad or self.b.requires_grad
        if training:
            if requires_grad:
                if self.w.requires_grad:
                    depends_on.append(Dependency(self.w, self.grad_w_dense))
                if self.b.requires_grad:
                    depends_on.append(Dependency(self.b, self.grad_b_dense))
                if x.requires_grad:
                    depends_on.append(Dependency(x, self.grad_a_dense))
            else:
                depends_on = []
        return self.type(data=data,requires_grad=requires_grad,depends_on=depends_on)


    def grad_w_dense(self, accum_grad):
        # Calculate gradient w.r.t layer weights
        grad_w = self.layer_input.T.dot(accum_grad)
        return grad_w

    def grad_b_dense(self, accum_grad):
        # Calculate gradient w.r.t layer weights
        grad_w0 = np.sum(accum_grad, axis=0, keepdims=True)
        return grad_w0

    def grad_a_dense(self, accum_grad):
        W = self.w.data
        accum_grad = accum_grad.dot(W.T)
        return accum_grad

    def update_pass(self):
        # Update the layer weights
        if self.trainable:
            self.w = self.w_opt.update(self.w)
            self.b = self.b_opt.update(self.b)
        # clear the gradients
        self.zero_grad()

    def output_shape(self):
        return (self.n_units, )



# Optimizers for models that use gradient based methods for finding the
# weights that minimizes the loss.
# A great resource for understanding these methods:
# http://sebastianruder.com/optimizing-gradient-descent/index.html

import numpy as np
from autograd.tensor import Tensor

class StochasticGradientDescent():
    __slots__ = ('learning_rate','momentum','w_updt')
    def __init__(self, learning_rate=0.01, momentum=0):
        self.learning_rate = learning_rate
        self.momentum = momentum
        self.w_updt = None

    def update(self, w):
        # If not initialized
        if self.w_updt is None:
            self.w_updt = np.zeros(np.shape(w.data))
        # Use momentum if set
        self.w_updt = w.grad  * (1 - self.momentum) + self.momentum * self.w_updt
        # Move against the gradient to minimize loss
        return w - self.learning_rate * self.w_updt

class Adam():
    __slots__ = ('learning_rate','eps','m','v','b1','b2','w_updt')
    def __init__(self, learning_rate=0.001, b1=0.9, b2=0.999):
        self.learning_rate = learning_rate
        self.eps = 1e-8
        self.m = None
        self.v = None
        # Decay rates
        self.b1 = b1
        self.b2 = b2

    def update(self, w):
        # If not initialized
        if self.m is None:
            self.m = np.zeros(np.shape(w.grad.data))
            self.v = np.zeros(np.shape(w.grad.data))

        self.m = w.grad * (1 - self.b1) + self.b1 * self.m
        self.v = (w.grad * w.grad) * (1 - self.b2) + self.b2 * self.v

        m_hat = self.m / (1 - self.b1)
        v_hat = self.v / (1 - self.b2)

        self.w_updt =  m_hat * self.learning_rate  / (np.sqrt(v_hat.data) + self.eps)

        return w - self.w_updt




import numpy as np
from autograd.dependency import Dependency


class Sum(object):
    __slots__ = ('type','t','axis','name','t_shape')

    def __init__(self, t, axis, name_child=None):
        self.type = type(t)
        self.t = t
        self.axis = axis
        self.name = name_child
        self.t_shape = None


    def _sum(self):
        #Takes a Tensor and returns the 0-Tensor that's the sum of all its elements.
        if self.axis == None:
            data = self.t.data.sum()
            self.t_shape = self.t.shape
        else:
            data = self.t.data.sum(axis=self.axis)
            self.t_shape = data.shape

        requires_grad = self.t.requires_grad
        if requires_grad:
            depends_on = [Dependency(self.t, self.grad_sum)]
        else:
            depends_on = []
        return self.type(data, requires_grad, depends_on, name=self.name)

    def grad_sum(self, grad: np.ndarray) -> np.ndarray:
        #grad is necessarily a 0-Tensor, so each input element contributes that much
        return grad * np.ones(self.t_shape)


class Add(object):
    __slots__ = ('type','t1','t2','name')

    def __init__(self, t1, t2, name_child=None):
        self.type = type(t1)
        self.t1 = t1
        self.t2 = t2
        self.name = name_child

    def _add(self):
        data = self.t1.data + self.t2.data
        requires_grad = self.t1.requires_grad or self.t2.requires_grad
        depends_on: List[Dependency] = []

        if self.t1.requires_grad:
            depends_on.append(Dependency(self.t1, self.grad_add1))
        if self.t2.requires_grad:
            depends_on.append(Dependency(self.t2, self.grad_add2))
        return self.type(data, requires_grad, depends_on, name=self.name)


    def grad_add1(self, grad: np.ndarray) -> np.ndarray:
        # Sum out added dims
        ndims_added = grad.ndim - self.t1.data.ndim
        for _ in range(ndims_added):
            grad = grad.sum(axis=0)
        # Sum across broadcasted (but non-added dims)
        for i, dim in enumerate(self.t1.shape):
            if dim == 1:
                grad = grad.sum(axis=i, keepdims=True)
        return grad

    def grad_add2(self, grad: np.ndarray) -> np.ndarray:
        # Sum out added dims
        ndims_added = grad.ndim - self.t2.data.ndim
        for _ in range(ndims_added):
            grad = grad.sum(axis=0)
        # Sum across broadcasted (but non-added dims)
        for i, dim in enumerate(self.t2.shape):
            if dim == 1:
                grad = grad.sum(axis=i, keepdims=True)
        return grad

class Mul(object):
    __slots__ = ('type','t1','t2','name')

    def __init__(self, t1, t2, name_child=None):
        self.type = type(t1)
        self.t1 = t1
        self.t2 = t2
        self.name = name_child

    def _mul(self):
        data = self.t1.data * self.t2.data
        requires_grad = self.t1.requires_grad or self.t2.requires_grad
        depends_on: List[Dependency] = []

        if self.t1.requires_grad:
            depends_on.append(Dependency(self.t1, self.grad_mul1))
        if self.t2.requires_grad:
            depends_on.append(Dependency(self.t2, self.grad_mul2))
        return self.type(data, requires_grad, depends_on, name=self.name)

    def grad_mul1(self, grad: np.ndarray) -> np.ndarray:
        grad = grad * self.t2.data
        # Sum out added dims
        ndims_added = grad.ndim - self.t1.data.ndim
        for _ in range(ndims_added):
            grad = grad.sum(axis=0)
        # Sum across broadcasted (but non-added dims)
        for i, dim in enumerate(self.t1.shape):
            if dim == 1:
                grad = grad.sum(axis=i, keepdims=True)
        return grad

    def grad_mul2(self, grad: np.ndarray) -> np.ndarray:
        grad = grad * self.t1.data
        # Sum out added dims
        ndims_added = grad.ndim - self.t2.data.ndim
        for _ in range(ndims_added):
            grad = grad.sum(axis=0)
        # Sum across broadcasted (but non-added dims)
        for i, dim in enumerate(self.t2.shape):
            if dim == 1:
                grad = grad.sum(axis=i, keepdims=True)
        return grad


class Neg(object):
    __slots__ = ('type','t','name')

    def __init__(self, t, name_child=None):
        self.type = type(t)
        self.t = t
        self.name = name_child

    def _neg(self):
        data = -self.t.data
        requires_grad = self.t.requires_grad
        if requires_grad:
            depends_on = [Dependency(self.t, lambda x: -x)]
        else:
            depends_on = []
        return self.type(data, requires_grad, depends_on, name=self.name)

class Sub(object):
    __slots__ = ('type','t1','t2','name')

    def __init__(self, t1, t2, name_child=None):
        self.type = type(t1)
        self.t1 = t1
        self.t2 = t2
        self.name = name_child

    def _sub(self):
        t3 = self.t1 + -self.t2
        t3.name = self.name
        return t3

class MatMul(object):
    __slots__ = ('type','t1','t2','name')

    def __init__(self, t1, t2, name_child = None):
        self.type = type(t1)
        self.t1 = t1
        self.t2 = t2
        self.name = name_child


    def _matmul(self):
        """
        if t1 is (n1, m1) and t2 is (m1, m2), then t1 @ t2 is (n1, m2)
        so grad3 is (n1, m2)
        if t3 = t1 @ t2, and grad3 is the gradient of some function wrt t3, then
            grad1 = grad3 @ t2.T
            grad2 = t1.T @ grad3
        """
        data = self.t1.data.dot(self.t2.data)
        requires_grad = self.t1.requires_grad or self.t2.requires_grad
        depends_on: List[Dependency] = []

        if self.t1.requires_grad:
            depends_on.append(Dependency(self.t1, self.grad_mm1))
        if self.t2.requires_grad:
            depends_on.append(Dependency(self.t2, self.grad_mm2))
        return self.type(data, requires_grad, depends_on, name=self.name)

    def grad_mm1(self, grad: np.ndarray) -> np.ndarray:
        return grad.dot(self.t2.data.T)

    def grad_mm2(self, grad: np.ndarray) -> np.ndarray:
        return self.t1.data.T.dot(grad)

class Div(object):
    __slots__ = ('type','t1','t2','name')

    def __init__(self, t1, t2, name_child = None):
        self.type = type(t1)
        self.t1 = t1
        self.t2 = t2
        self.name = name_child

    def _div(self):
        data = self.t1.data / self.t2.data
        requires_grad = self.t1.requires_grad or self.t2.requires_grad
        depends_on: List[Dependency] = []
        if self.t1.requires_grad:
            depends_on.append(Dependency(self.t1, self.grad_div1))
        if self.t2.requires_grad:
            depends_on.append(Dependency(self.t2, self.grad_div2))
        return self.type(data, requires_grad, depends_on, name=self.name)

    def grad_div1(self, grad: np.ndarray) -> np.ndarray:
        grad = grad * (1 / self.t2.data)
        # Sum out added dims
        ndims_added = grad.ndim - self.t1.data.ndim
        for _ in range(ndims_added):
            grad = grad.sum(axis=0)
        # Sum across broadcasted (but non-added dims)
        for i, dim in enumerate(self.t1.shape):
            if dim == 1:
                grad = grad.sum(axis=i, keepdims=True)
        return grad

    def grad_div2(self, grad: np.ndarray) -> np.ndarray:
        grad = grad * (-self.t1.data / (self.t2.data * self.t2.data))
        # Sum out added dims
        ndims_added = grad.ndim - self.t2.data.ndim
        for _ in range(ndims_added):
            grad = grad.sum(axis=0)
        # Sum across broadcasted (but non-added dims)
        for i, dim in enumerate(self.t2.shape):
            if dim == 1:
                grad = grad.sum(axis=i, keepdims=True)
        return grad



class Exp(object):
    __slots__ = ('type','t','name')

    """docstring for Exp."""
    def __init__(self, t,  name_child = None):
        self.type = type(t)
        self.t = t
        self.name = name_child

    def _exp(self):
        data = np.exp(self.t.data)
        requires_grad = self.t.requires_grad
        if requires_grad:
            depends_on = [Dependency(self.t, self.grad_exp)]
        else:
            depends_on = []
        return self.type(data, requires_grad, depends_on, name = self.name)

    def grad_exp(self, grad: np.array):
        grad = grad * np.exp(self.t.data)
        # Sum out added dims
        ndims_added = grad.ndim - self.t.data.ndim
        for _ in range(ndims_added):
            grad = grad.sum(axis=0)
        # Sum across broadcasted (but non-added dims)
        for i, dim in enumerate(self.t.shape):
            if dim == 1:
                grad = grad.sum(axis=i, keepdims=True)
        return grad


class Max(object):
    __slots__ = ('type','t','name')

    """docstring for Max."""
    def __init__(self, t, name_child = None):
        self.type = type(t)
        self.t = t
        self.name = name_child

    def _max(self):
        data = self.t.data.max()
        requires_grad = self.t.requires_grad
        if requires_grad:
            depends_on = [Dependency(self.t, self.grad_max)]
        else:
            depends_on = []
        return self.type(data, requires_grad, depends_on, name=self.name)

    def grad_max(self, grad: np.array):
        grad = np.equal(self.t.data, self.t.data.max()) * grad.sum()
        # Sum out added dims
        ndims_added = grad.ndim - self.t.data.ndim
        for _ in range(ndims_added):
            grad = grad.sum(axis=0)
        # Sum across broadcasted (but non-added dims)
        for i, dim in enumerate(self.t.shape):
            if dim == 1:
                grad = grad.sum(axis=i, keepdims=True)
        return grad



import torch
import torchvision.models as models
import numpy as np

class Detach_params(object):
    """docstring for Detach_params."""

    def __init__(self, model=None):
        super(Detach_params, self).__init__()
        self.model = model
        self.modelparams = {}
        self.modelfeatures = {}

    def show_model(self):
        print(self.model)

    def detach_params(self):
        for name, param in self.model.named_parameters():
            self.modelparams[str(name)] = param.detach().numpy().astype("float32")

    def save_params(self, path, name):
        np.save(path+name, self.modelparams)

    def load_params(self, path, name):
        return np.load(path+name,allow_pickle=True)

    def detach_features(self):
        for name, param in self.model.features.named_parameters():
            self.modelparams[str(name)] = param.detach().numpy().astype("float32")



import cupy as cp
from autogradgpu.tensor import Tensor
from autogradgpu.parameter import Parameter
from autogradgpu.dependency import Dependency
import math
import copy
from metalgpu.layers.layer import Layer


class Dropout(Layer):
    __slots__ =( 'p', '_mask','n_units', 'pass_through','trainable','type' )

    """A layer that randomly sets a fraction p of the output units of the previous layer
    to zero.
    Parameters:
    -----------
    p: float
        The probability that unit x is set to zero.
    """
    def __init__(self, p=0.2):
        self.p = p
        self._mask = None
        self.input_shape = None
        self.n_units = None
        self.pass_through = True
        self.trainable = True

    def forward_pass(self, X, training=True):
        requires_grad =  X.requires_grad
        self.type = type(X)
        c = (1 - self.p)
        if training:
            self._mask = cp.random.uniform(size=X.shape) > self.p
            c = self._mask
        data =  X.data * c

        if requires_grad:
            depends_on = [Dependency(X, self.grad_dropout)]
        else:
            depends_on = []
        return self.type(data=data,requires_grad=requires_grad,depends_on=depends_on)

    def grad_dropout(self, accum_grad):
        return accum_grad * self._mask

    def update_pass(self):
        # clear the gradients
        self.zero_grad()

    def output_shape(self):
        return self.input_shape



import resource
import sys
import urllib.request
import PIL
from PIL import Image
from metalgpu.utils.data_manipulation import normalize
import dill
from autogradgpu.parameter import Parameter
import random
import cv2
import os
import scipy.ndimage as ndi

import cupy as cp
import scipy
import matplotlib.pyplot as plt
import seaborn as sns
from scipy.ndimage import rotate

sns.set(color_codes=True)

print('resource limit ',resource.getrlimit(resource.RLIMIT_STACK))
print('recursion limit ',sys.getrecursionlimit())


def save_model(filename,model, max_rec=100000):
    max_rec = max_rec
    sys.setrecursionlimit(max_rec)
    with open(filename+'.pkl', 'wb') as file:
        dill.dump(model.layers, file)

def load_model(filename):
    with open(filename+'.pkl', 'rb') as file:
        layers = dill.load(file)
    return layers

def fetch_img(url,shape,invert=False):
    I = Image.open(urllib.request.urlopen(url)).convert('L')
    I = cp.asarray(I.resize((shape[-1],shape[-2]), PIL.Image.LANCZOS)).reshape(*shape)
    if invert:
        I = cp.invert(I)
    I = normalize(I)
    return I

def _forward_pass(X, training=True, model_name=None):
    """ Calculate the output of the NN """
    layers_ = load_model(model_name)
    layer_output = Parameter(X,False)
    for layer in layers_:
        layer_output = layer.forward_pass(layer_output, training)
    return layer_output

#resize batch
#example: img_tran(X_train,8)
def imgs_trans(imgs_in,size):
    factor = size/imgs_in.shape[1]
    imgs_out = ndi.zoom(imgs_in, (1, factor, factor, 1), order=2)
    print(imgs_out.shape)
    return imgs_out

def create_training_data(img_size=50, classes=[], data_dir="", color=None):
    #classses is the name of each folder in downloads
    #data_dir is path to downloads
    #img_size for WxH

    if color == 'gray':
        color = cv2.IMREAD_GRAYSCALE
    elif color == 'gray_to_rgb':
        color = cv.CV_GRAY2RGB
    elif color == None:
        color = cv2.COLOR_BGR2RGB

        # TODO: add more opitons

    training_data = []
    for class_ in classes:
        path = os.path.join(data_dir, class_)
        class_num = classes.index(class_)
        for img in os.listdir(path):
            try:
                img_array = cv2.imread(os.path.join(path,img), color)
                new_array = cv2.resize(img_array,(img_size,img_size)) #bug need to fix channels
                training_data.append([new_array,class_num])
            except Exception  as e:
                pass
    random.shuffle(training_data)
    imgs = []
    labs = []
    for img, lab in training_data:
        imgs.append(img)
        labs.append(lab)
    #imgs = cp.array(imgs).reshape(-1, img_size,img_size)
    self.imgs = imgs
    self.labels = labels
    return self.imgs, self.labs



import numpy as np
from autograd.tensor import Tensor
from autograd.parameter import Parameter
from metal.module import Module
from autograd.dependency import Dependency
import math
import copy
from metal.layers.layer import Layer

class Flatten(Layer):
    __slots__ =( 'prev_shape', 'trainable','type' )

    """ Turns a multidimensional matrix into two-dimensional """
    def __init__(self, input_shape=None):
        self.prev_shape = None
        self.trainable = True
        self.input_shape = input_shape

    def forward_pass(self, x_, training=True):
        self.prev_shape = x_.shape
        X = x_.data
        self.type = type(x_)
        requires_grad = x_.requires_grad
        if training:
            if requires_grad:
                depends_on = [Dependency(x_, self.gardflatten)]
            else:
                depends_on = []
        else:
            depends_on = []
        return self.type(data=X.reshape((X.shape[0], -1)),requires_grad=requires_grad,depends_on=depends_on)

    def gardflatten(self, accum_grad):
        return accum_grad.reshape(self.prev_shape)

    def update_pass(self):
        pass

    def output_shape(self):
        return (np.prod(self.input_shape),)
